###################################################

from os import environ

import certifi
import urllib3

from http.client import responses

from getpass import getpass

import random, string

###################################################

http = urllib3.PoolManager(
    cert_reqs = 'CERT_REQUIRED',
    ca_certs = certifi.where()
)

###################################################

API_URL = "https://lichess.org/api"
MAX_TOURNEY_LENGTH_MINS = 360

###################################################

def getalltourneysurl():
    return "{}/tournament".format(API_URL)

def gettourneyurl(tid):
    return "{}/tournament/{}".format(API_URL, tid)

def gettourneygamesurl(tid):
    return "{}/tournament/{}/games".format(API_URL, tid)

###################################################

def randsri():
    return ''.join(random.choice(string.ascii_lowercase + string.digits) for _ in range(10))

def tourneychaturl(tid):    
    return "wss://socket.lichess.org/tournament/{}/socket/v4?sri={}".format(tid, randsri())

#########################################################

FEATURED_TOURNEY_NAME = environ.get("FEATUREDTOURNEY", "AtomicChessBot Blitz Tourney")

def fulltourneyname(tourneyname):
    return tourneyname + " Arena"

FULL_FEATURED_TOURNEY_NAME = fulltourneyname(FEATURED_TOURNEY_NAME)

FEATURED_TOURNEY_CREATOR = environ.get("FEATUREDTOURNEYCREATOR", "handywebprojects")

CHAT_USER = environ.get("CHATUSER", "UltraBulletBot")
CHAT_PASS = environ.get("CHATPASS", None)

###################################################

def login(username = None, password = None):
    if not username:
        username = input("Username: ")
    if not password:
        password = getpass()

    res = http.request("POST", "https://lichess.org/login?referrer=%2F", headers = {
        "Referer": "https://lichess.org/login?referrer=%2F"
    }, fields = {
        "username": username,
        "password": password
    })

    print("login status", responses[res.status])

    if res.status == 200:
        sch = res.info().get("Set-Cookie")

        lila2 = sch.split(";")[0][6:]

        print("for {} obtained: {}".format(username, lila2))
        return lila2
    else:
        print("could not log in with {}".format(username))
        return None

###################################################

def chatuserlila2():
    return login(CHAT_USER, CHAT_PASS)


###################################################

import threading, time

import websocket

from queue import Queue
import json

import requests

###################################################

from obscureridge import lichess

###################################################

KEEPALIVE_SLEEP = 3
MAX_KEEPALIVE_SECS = lichess.MAX_TOURNEY_LENGTH_MINS * 60

###################################################

class Message:
    def __init__(self, kind, data):
        self.kind = kind
        self.data = data

    def __repr__(self):
        return "{}".format(self.__dict__)

class Chatbot:
    def __init__(self,
        lila2,
        tid
    ):
        self.lila2 = lila2
        self.tid = tid
        self.tourneychaturl = lichess.tourneychaturl(self.tid)
        print("initialized chatbot {} {}".format(self.tid, self.tourneychaturl))

    def say_thread_target(self, msg):
        self.ws.send(json.dumps({
            "t": "talk",
            "d": msg
        }))
        print("said {} : {}".format(self.tid, msg))

    def say(self, msg):
        threading.Thread(target = self.say_thread_target, args = (msg,)).start()

    def talkhandler(self, user, msg):
        print("message {} : {}".format(user, msg))

    def crowdhandler(self, nb, users, anons):
        print("crowd {} : users = {} , anons = {}".format(nb, " ".join(users), anons))

    def gamefinishedhandler(self, gid):        
        gobj = self.games[gid]
        print("game finished {} : {} ( {} {} ) - {} ( {} {} ) {} \n {}".format(
            gid,
            gobj["whiteName"],
            gobj["whiteRating"],
            gobj["whiteRatingDiff"],
            gobj["blackName"],
            gobj["blackRating"],
            gobj["blackRatingDiff"],
            gobj["score"],
            gobj["moves"])
        )

    def messagehandler_thread_target(self):
        while self.alive:            
            try:
                message = self.messagequeue.get(timeout = KEEPALIVE_SLEEP)                                
                if message.kind == "message":
                    self.talkhandler(message.data["u"], message.data["t"])
                elif message.kind == "crowd":
                    self.crowdhandler(message.data.get("nb", 0), message.data.get("users", []), message.data.get("anons", 0))
                elif message.kind == "gamefinished":
                    self.gamefinishedhandler(message.data)
            except:
                pass
        print("message handler thread terminated for {}".format(self.tid))

    def keepalive_thread_target(self):                
        while self.keepalivecnt > 0:            
            self.ws.send("null")
            time.sleep(KEEPALIVE_SLEEP)
            self.keepalivecnt -= 1
        self.ws.close()
        print("keep alive thread terminated for {}".format(self.tid))

    def read_games_thread_target(self):
        first = True
        while self.alive:
            r = requests.get(lichess.gettourneygamesurl(self.tid), headers = {"Accept": "application/x-ndjson"}, stream = True)
            for line in r.iter_lines():
                try:
                    line = line.decode("utf-8")                    
                    gobj = json.loads(line)
                    gid = gobj["id"]
                    status = gobj["status"]                    
                    if not ( status == "started" ):
                        if not ( gid in self.gameids ):
                            self.gameids.append(gid)
                            players = gobj.get("players", {})
                            for color in ["white", "black"]:
                                player = players.get(color, {})
                                rating = player.get("rating", 1500)
                                ratingdiff = player.get("ratingDiff", 0)
                                user = player.get("user", {})
                                gobj[color + "Name"] = user.get("name")
                                gobj[color + "Rating"] = rating
                                gobj[color + "RatingDiff"] = ratingdiff
                                winner = gobj.get("winner", None)
                                gobj["winner"] = winner
                                gobj["score"] = 0                                
                                if winner:
                                    if winner == "white":
                                        gobj["score"] = 1
                                    else:
                                        gobj["score"] = -1
                            gobj["ratingBias"] = gobj["whiteRating"] - gobj["blackRating"]
                            moves = gobj.get("moves", "")
                            if moves == "":
                                gobj["moves"] = []
                            else:
                                gobj["moves"] = moves.split(" ")
                            self.games[gid] = gobj                            
                            if ( not first ):
                                self.messagequeue.put(Message("gamefinished", gid))
                        else:
                            r.close()
                            break
                except:
                    pass            
            first = False
            time.sleep(5 * KEEPALIVE_SLEEP)
        print("read games thread terminated for {}".format(self.tid))

    def shutdown(self):
        self.keepalivecnt = 0

    def on_open(self):
        print("socket for {} opened".format(self.tid))
        threading.Thread(target = self.keepalive_thread_target).start()
        threading.Thread(target = self.messagehandler_thread_target).start()
        threading.Thread(target = self.read_games_thread_target).start()

    def on_message(self, message):
        try:
            mobj = json.loads(message)
            kind = mobj["t"]
            data = mobj["d"]            
            self.messagequeue.put(Message(kind, data))
        except:
            pass

    def startup_thread_target(self):
        print("starting up chatbot for {}".format(self.tid))

        self.keepalivecnt = int(MAX_KEEPALIVE_SECS / KEEPALIVE_SLEEP)        

        self.alive = True

        self.messagequeue = Queue()

        self.gameids = []
        self.games = {}

        self.ws = websocket.WebSocketApp(self.tourneychaturl,        
            on_open = self.on_open,        
            on_message = self.on_message,
            cookie = "lila2={}".format(self.lila2)
        )

        self.ws.run_forever(
            host = "socket.lichess.org",
            origin = "https://lichess.org"
        )

        print("socket terminated for {}".format(self.tid))

        self.alive = False

    def startup(self):
        threading.Thread(target = self.startup_thread_target).start()

###################################################


###################################################

import certifi
import urllib3

import json

from traceback import print_exc as pe

###################################################

http = urllib3.PoolManager(
    cert_reqs = 'CERT_REQUIRED',
    ca_certs = certifi.where()
)

###################################################

def geturl(url, headers = {}, encoding = "utf-8"):        
    try:
        print("getting", url)
        r = http.request("GET", url, headers = headers)
        content = r.data.decode(encoding)
        print("received length", len(content), "for", url)
    except:
        pe()
        content = None
    return content

def geturljson(url, headers = {}):            
    try:
        content = geturl(url, headers)
        obj = json.loads(content)        
    except:
        pe()
        obj = None
    return obj

###################################################



###################################################

from obscureridge import lichess
from obscureridge.urllibutils import geturljson
from obscureridge.chatbot import Chatbot

from os import environ
import time
import random

###################################################

def getallfeaturedtourneys(kinds = ["created", "started"]):
    alltourneys = geturljson(lichess.getalltourneysurl())    
    allfeaturedtourneys = []
    for kind in kinds:
        tourneys = alltourneys[kind]
        for tourney in tourneys:
            createdby = tourney["createdBy"]
            name = tourney["fullName"]
            if ( createdby == lichess.FEATURED_TOURNEY_CREATOR ) and ( name == lichess.FULL_FEATURED_TOURNEY_NAME ):
                allfeaturedtourneys.append(tourney)
    return allfeaturedtourneys

###################################################

class MyChatbot(Chatbot):
    def __init__(self, lila2, tid):
        super().__init__(lila2, tid)

    def talkhandler(self, user, msg):
        print("MY message {} : {}".format(user, msg))

    def crowdhandler(self, nb, users, anons):
        print("MY crowd {} : users = {} , anons = {}".format(nb, " ".join(users), anons))

    def gamefinishedhandler(self, gid):                
        gobj = self.games[gid]
        ratingbias = gobj["ratingBias"]
        score = gobj["score"]
        nomoves = len(gobj["moves"])
        whitename = gobj["whiteName"]
        whiterating = gobj["whiteRating"]
        blackname = gobj["blackName"]
        blackrating = gobj["blackRating"]
        rs = ratingbias * score        
        if ( rs < -100 ) and ( nomoves > 1 ):            
            upset = whitename            
            upsetloser = blackname
            if score < 0:
                upset = blackname
                upsetloser = whitename            
            self.say(random.choice([
                "what a game {} against {}".format(upset, upsetloser),
                "wow {} defeats {}".format(upset, upsetloser),
                "unbelievable {} beating {} higher rated".format(upset, abs(ratingbias))
            ]))
            if upset == blackname:
                self.say("with black!")            

def startup():
    chatuserlila2 = environ.get("CHATUSERLILA2", None)

    tourneys = getallfeaturedtourneys()

    if len(tourneys) > 0:
        tourney = tourneys[0]
        tid = tourney["id"]
        print("found tourney {}".format(tid))
    else:
        print("could not find tourney")

    chatbot = MyChatbot(chatuserlila2, tid)

    chatbot.startup()

    time.sleep(3600)

    chatbot.shutdown()

###################################################


from setuptools import setup
from obscureridge import read_string_from_file

setup(name='obscureridge',
      version='0.0.3',
      author='pythonflaskserverapps',
      author_email='pythonflaskserverapps@gmail.com',
      description='test project',
      long_description=read_string_from_file("README.md", "Test project."),
      long_description_content_type='text/markdown',
      license='MIT',
      keywords="test project",
      url='https://github.com/pythonflaskserverapps/obscureridge',            
      packages=['obscureridge'],
      test_suite="travis_test",
      python_requires=">=3.6",
      install_requires=[                
      ],
      classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",        
        "Programming Language :: Python :: 3.6"
      ],
      entry_points={
        'console_scripts': []
      },
      package_dir={
        'obscureridge': 'obscureridge'
      },
      include_package_data=False,
      zip_safe=False)

